{
  "type": "record",
  "name": "DataPlatformInfo",
  "namespace": "com.linkedin.pegasus2avro.dataplatform",
  "doc": "Information about a data platform",
  "fields": [
    {
      "name": "name",
      "type": "string",
      "doc": "Name of the data platform",
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": false,
        "fieldType": "TEXT_PARTIAL"
      },
      "validate": {
        "strlen": {
          "max": 15
        }
      }
    },
    {
      "name": "displayName",
      "type": [
        "null",
        "string"
      ],
      "doc": "The name that will be used for displaying a platform type.",
      "default": null,
      "Searchable": {
        "boostScore": 10.0,
        "enableAutocomplete": true,
        "fieldType": "TEXT_PARTIAL"
      }
    },
    {
      "name": "type",
      "type": {
        "type": "enum",
        "name": "PlatformType",
        "doc": "Platform types available at LinkedIn",
        "symbols": [
          "FILE_SYSTEM",
          "KEY_VALUE_STORE",
          "MESSAGE_BROKER",
          "OBJECT_STORE",
          "OLAP_DATASTORE",
          "OTHERS",
          "QUERY_ENGINE",
          "RELATIONAL_DB",
          "SEARCH_ENGINE"
        ],
        "symbolDocs": {
          "FILE_SYSTEM": "Value for a file system, e.g. hdfs",
          "KEY_VALUE_STORE": "Value for a key value store, e.g. espresso, voldemort",
          "MESSAGE_BROKER": "Value for a message broker, e.g. kafka",
          "OBJECT_STORE": "Value for an object store, e.g. ambry",
          "OLAP_DATASTORE": "Value for an OLAP datastore, e.g. pinot",
          "OTHERS": "Value for other platforms, e.g salesforce, dovetail",
          "QUERY_ENGINE": "Value for a query engine, e.g. presto",
          "RELATIONAL_DB": "Value for a relational database, e.g. oracle, mysql",
          "SEARCH_ENGINE": "Value for a search engine, e.g seas"
        }
      },
      "doc": "Platform type this data platform describes"
    },
    {
      "name": "datasetNameDelimiter",
      "type": "string",
      "doc": "The delimiter in the dataset names on the data platform, e.g. '/' for HDFS and '.' for Oracle"
    },
    {
      "name": "logoUrl",
      "type": [
        "null",
        "string"
      ],
      "doc": "The URL for a logo associated with the platform",
      "default": null,
      "java": {
        "class": "com.linkedin.pegasus2avro.common.url.Url",
        "coercerClass": "com.linkedin.pegasus2avro.common.url.UrlCoercer"
      }
    }
  ],
  "Aspect": {
    "name": "dataPlatformInfo"
  }
}